/* gtk-3dcanvas-ball - a ball to put on a gtk-3dcanvas */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/* Useful */
#define PI                      3.14159265358979323846264338327
#define DEG(angleRad)           ((angleRad) * 180 / PI)
#define RAD(angle360)           ((angle360) * PI / 180)
#define SQR(x)                  ((x) * (x))
#define SIGN(x)                 (((x) >= 0) ? 1.0 : -1.0)

/*---------------------------------\
| Property and signals definitions |
\---------------------------------*/
enum {
        BALL_PROP_0,
	BALL_PROP_CENTER_X,
	BALL_PROP_CENTER_Y,
	BALL_PROP_CENTER_Z,
	BALL_PROP_RADIUS,
	BALL_PROP_SMOTHNESS
};
/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_ball_class_init (Gtk3dcanvasBallClass *class);
static void gtk_3dcanvas_ball_init(Gtk3dcanvasBall *glb);

static GObject *
gtk_3dcanvas_ball_constructor (GType             type,
			       guint             n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_ball_finalize (GObject *object);
static void
gtk_3dcanvas_ball_dispose (GObject *object);

static void
gtk_3dcanvas_ball_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec);
static void
gtk_3dcanvas_ball_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
                                GParamSpec           *pspec);

static void
_gtk_3dcanvas_ball_add_ball(Gtk3dcanvasBall *ball);

static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_ball_get_type]-----------------------------------------*/
GType gtk_3dcanvas_ball_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasBallClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_ball_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasBall),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_ball_init,
    };

    type = g_type_register_static (gtk_3dcanvas_item_get_type (),
				   "Gtk3dcanvasBall",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_ball_class_init ]------------------------------------*/
static void gtk_3dcanvas_ball_class_init (Gtk3dcanvasBallClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->constructor  = gtk_3dcanvas_ball_constructor;
  object_class->set_property = gtk_3dcanvas_ball_set_property;
  object_class->get_property = gtk_3dcanvas_ball_get_property;
  object_class->finalize     = gtk_3dcanvas_ball_finalize;
  object_class->dispose      = gtk_3dcanvas_ball_dispose;

  g_object_class_install_property
    (object_class, BALL_PROP_CENTER_X,
     g_param_spec_float ("center_x", 
			 _("Center X"), 
			 _("X coordinate of the center"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, BALL_PROP_CENTER_Y,
     g_param_spec_float ("center_y", 
			 _("Center Y"), 
			 _("Y coordinate of the center"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, BALL_PROP_CENTER_Z,
     g_param_spec_float ("center_z", 
			 _("Center Z"), 
			 _("Z coordinate of the center"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, BALL_PROP_RADIUS,
     g_param_spec_float ("radius", 
			 _("Radius"), 
			 _("Radius of the ball"),
			 0,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, BALL_PROP_SMOTHNESS,
     g_param_spec_int ("smoothness", 
		       _("Smoothness"), 
		       _("Smoothness of the ball. The bigger it is, the more points the ball will contain. About 10 is really ok."),
		       0,G_MAXINT,0,
		       (G_PARAM_READWRITE)));

}

/*----[ gl_canvas_ball_constructor ]--------------------------------------*/
static GObject *
gtk_3dcanvas_ball_constructor (GType             type,
				guint            n_construct_properties,
			  GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_ball_finalize ]--------------------------------------*/
static void
gtk_3dcanvas_ball_finalize (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_ball_dispose ]------------------------------------------*/
static void
gtk_3dcanvas_ball_dispose (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->dispose (object);
}

/*----[ gtk_3dcanvas_ball_init ]---------------------------------------------*/
static void gtk_3dcanvas_ball_init(Gtk3dcanvasBall *ball) {
} 

/*----[ gtk_3dcanvas_ball_set_property ]-------------------------------------*/
static void
gtk_3dcanvas_ball_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec) {
  Gtk3dcanvasBall *ball;
  Gtk3dcanvasItem *item;
  float delta;
  int i;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_BALL (object));

  ball = GTK_3DCANVAS_BALL(object);
  item = GTK_3DCANVAS_ITEM (object);

  switch (param_id) {
  case BALL_PROP_CENTER_X:
    delta = ball->center.x - g_value_get_float(value);
    ball->center.x = g_value_get_float(value);
    for (i=0;i<item->iarrays->len;i++)
      g_array_index(item->iarrays,Gtk3dOPoint,i).x += delta;
    break;
  case BALL_PROP_CENTER_Y:
    delta = ball->center.y - g_value_get_float(value);
    ball->center.y = g_value_get_float(value);
    for (i=0;i<item->iarrays->len;i++)
      g_array_index(item->iarrays,Gtk3dOPoint,i).y += delta;
    break;
  case BALL_PROP_CENTER_Z:
    delta = ball->center.z - g_value_get_float(value);
    ball->center.z = g_value_get_float(value);
    for (i=0;i<item->iarrays->len;i++)
      g_array_index(item->iarrays,Gtk3dOPoint,i).z += delta;
    break;

  case BALL_PROP_RADIUS:
    /* FIXME: could be done with transformation on points without freeing anything */
    gtk_3dcanvas_item_empty(item);
    ball->radius=g_value_get_float(value);
    _gtk_3dcanvas_ball_add_ball(ball);
    break;
  case BALL_PROP_SMOTHNESS:
    gtk_3dcanvas_item_empty(item);
    ball->smoothness=g_value_get_int(value);
    _gtk_3dcanvas_ball_add_ball(ball);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
  gtk_3dcanvas_draw(item->canvas);
}

/*----[ gtk_3dcanvas_ball_get_property ]-------------------------------------*/
static void
gtk_3dcanvas_ball_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
				GParamSpec           *pspec) {
  Gtk3dcanvasItem *item;
  Gtk3dcanvasBall *ball;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_BALL (object));

  ball = GTK_3DCANVAS_BALL(object);
  item = GTK_3DCANVAS_ITEM (object);

  switch (param_id) {
  case BALL_PROP_CENTER_X:
    g_value_set_float (value, ball->center.x);
    break;

  case BALL_PROP_CENTER_Y:
    g_value_set_float (value, ball->center.y);
    break;

  case BALL_PROP_CENTER_Z:
    g_value_set_float (value, ball->center.z);
    break;

  case BALL_PROP_RADIUS:
    g_value_set_float (value, ball->radius);
    break;

  case BALL_PROP_SMOTHNESS:
    g_value_set_int (value, ball->smoothness);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
}

/*----[ _gtk_3dcanvas_ball_add_ball ]----------------------------------------*/
static void
_gtk_3dcanvas_ball_add_ball(Gtk3dcanvasBall *ball) {
  
  float x0,y0,z0;
  float radius;
  int smoothness;
  Gtk3dcanvasItem *item;

  float theta;
  float phi;
  float x, y, z;
  int p, t;
  int range_t, range_p;
  int center_v, pole_v, begin_v = 0;
  int obj_v;

  g_return_if_fail (ball != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_BALL (ball));
  
  item=GTK_3DCANVAS_ITEM(ball);
  x0 =ball->center.x;
  y0 =ball->center.y;
  z0 =ball->center.z;
  radius = ball->radius;
  smoothness = ball->smoothness;

  printf("Add ball: center= %.1f %.1f %.1f; radius=%.1f; smooth=%d\n",
	 x0,y0,z0,radius,smoothness);

  range_t = smoothness;
  range_p = 2 * smoothness;

  /* Center vertex */
  center_v = gtk_3dcanvas_item_add_point( item, x0, y0, z0, 0.0, 0.0, -1.0 );
  gtk_3dcanvas_item_add_index( item, center_v );
  gtk_3dcanvas_item_add_index( item, center_v );
  
  /* Front of ball (North Pole to Arctic Circle) */
  
  /* Pole vertex */
  pole_v = gtk_3dcanvas_item_add_point( item, x0 + radius, y0, z0, 1.0, 0.0, 0.0 );
  
  theta = RAD(90.0 - 180.0 / (float)range_t);
  x = radius * sin( theta );
  for (p = 0; p < range_p; p++) {
    phi = RAD(360.0 * (float)p / (float)range_p);
    y = radius * cos( theta ) * cos( phi );
    z = radius * cos( theta ) * sin( phi );
    obj_v = gtk_3dcanvas_item_add_point( item, x0 + x, y0 + y, z0 + z, x, y, z );
    gtk_3dcanvas_item_add_index( item, pole_v );
    gtk_3dcanvas_item_add_index( item, obj_v );
    if (p == 0)
      begin_v = obj_v;
  }
  
  /* Complete the circle */
  gtk_3dcanvas_item_add_index( item, pole_v );
  gtk_3dcanvas_item_add_index( item, begin_v );
  
  /* Midsection of ball (Arctic Circle to Antarctic Circle) */
  
  for (t = 2; t <  range_t; t++) {
    theta = RAD(90.0 - 180.0 * (float)t / (float)range_t);
    x = radius * sin( theta );
    for (p = 0; p < range_p; p++) {
      phi = RAD(360.0 * (float)p / (float)range_p);
      y = radius * cos( theta ) * cos( phi );
      z = radius * cos( theta ) * sin( phi );
      obj_v = gtk_3dcanvas_item_add_point( item, x0 + x, y0 + y, z0 + z, x, y, z );
      gtk_3dcanvas_item_add_index( item, obj_v - range_p );
      gtk_3dcanvas_item_add_index( item, obj_v );
      if (p == 0)
	begin_v = obj_v;
    }
    
    /* Complete the circle */
    gtk_3dcanvas_item_add_index( item, begin_v - range_p );
    gtk_3dcanvas_item_add_index( item, begin_v );
  }
  
  /* Back of ball (Antarctic Circle to South Pole) */
  
  /* Pole vertex */
  pole_v = gtk_3dcanvas_item_add_point( item, x0 - radius, y0, z0, -1.0, 0.0, 0.0 );
  
  theta = RAD(180.0 / (float)range_t - 90.0);
  x = radius * sin( theta );
  for (p = 0; p < range_p; p++) {
    phi = RAD(360.0 * (float)p / (float)range_p);
    y = radius * cos( theta ) * cos( phi );
    z = radius * cos( theta ) * sin( phi );
    obj_v = gtk_3dcanvas_item_add_point( item, x0 + x, y0 + y, z0 + z, x, y, z );
    gtk_3dcanvas_item_add_index( item, obj_v );
    gtk_3dcanvas_item_add_index( item, pole_v );
    if (p == 0)
      begin_v = obj_v;
  }
  
  /* Once more, complete the circle */
  gtk_3dcanvas_item_add_index( item, begin_v );
  gtk_3dcanvas_item_add_index( item, pole_v );
  
  /* and tie quad strip back to center */
  gtk_3dcanvas_item_add_index( item, center_v );
  gtk_3dcanvas_item_add_index( item, center_v );

}
/*-----------------\
| Public functions |
\-----------------*/
/**
 * gtk_3dcanvas_ball_new:
 * @parent: group in which to add the ball
 * @x0: x of the center point
 * @y0: y of the center point
 * @z0: z of the center point
 * @color: the color of the ball
 * @radius: radius of the ball
 * @smoothness: the bigger it is, the more point we will generate (10 is ok)
 * @Returns: the newly created ball.
 * 
 * Make a ball item. 
 **/
Gtk3dcanvasItem* gtk_3dcanvas_ball_new(Gtk3dcanvasGroup *parent,
				       float x0, float y0, float z0, 
				       char *color,
				       float radius, int smoothness ) {

  Gtk3dcanvasBall *ball;
  Gtk3dcanvasItem *item;
  
  item=gtk_3dcanvas_item_new(parent,GTK_3DCANVAS_BALL_TYPE,
			     "type",GL_QUAD_STRIP,
			     "color",color,
			     NULL);
  ball=GTK_3DCANVAS_BALL(item);

  /* Update the wannabe properties */
  ball->center.x    = x0;
  ball->center.y    = y0;
  ball->center.z    = z0;
  ball->radius     = radius;
  ball->smoothness = smoothness;

  /* add the ball */
  _gtk_3dcanvas_ball_add_ball(ball);
  return item;
}
