/* gtk-3dcanvas-group - collection of gtk-3dcanvas-item */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_group_class_init (Gtk3dcanvasGroupClass *class);
static void gtk_3dcanvas_group_init(Gtk3dcanvasGroup *glb);

static GObject *
gtk_3dcanvas_group_constructor (GType                  type,
				guint                  n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_group_finalize (GObject *object);

static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_group_get_type]---------------------------------------*/
GType gtk_3dcanvas_group_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasGroupClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_group_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasGroup),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_group_init,
    };

    type = g_type_register_static (gtk_3dcanvas_item_get_type (),
				   "Gtk3dcanvasGroup",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_group_class_init ]------------------------------------*/
static void gtk_3dcanvas_group_class_init (Gtk3dcanvasGroupClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  /* nothing to init in the class for now */
  object_class->constructor  = gtk_3dcanvas_group_constructor;
  //  object_class->set_property = gl_browser_set_property;
  //  object_class->get_property = gl_browser_get_property;
  object_class->finalize     = gtk_3dcanvas_group_finalize;

}

/*----[ gl_canvas_group_constructor ]----------------------------------------*/
static GObject *
gtk_3dcanvas_group_constructor (GType                  type,
				guint                  n_construct_properties,
			  GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_group_finalize ]--------------------------------------*/
static void
gtk_3dcanvas_group_finalize (GObject *object)
{
  //  Gtk3dcanvasGroup *group = GTK_3DCANVAS_GROUP (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_group_init ]------------------------------------------*/
static void gtk_3dcanvas_group_init(Gtk3dcanvasGroup *group) {
  group->childs = g_array_new(FALSE,FALSE,sizeof(Gtk3dcanvasGroup*));
} 

/*----------------------\
| Semi-Public functions |
\----------------------*/
#undef  MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#undef  MAX
#define MAX(a,b) ((a)>(b)?(a):(b))

void gtk_3dcanvas_group_compute_extend(Gtk3dcanvasGroup *group) {
  int i;
  Gtk3dcanvasItem *item;

  if (!group)
    return;

#ifdef DEBUG
  printf ("%d childs:\n",group->childs->len);
  for (i=0;i<group->childs->len;i++) {
    item=g_array_index(group->childs,Gtk3dcanvasItem*,i);
    printf("  %p extents %p x:%.2f-%.2f y:%.2f-%.2f z:%.2f-%.2f avg:%.2f\n",item,&(item->extents),
	   item->extents.xmin,item->extents.xmax,
	   item->extents.ymin,item->extents.ymax,
	   item->extents.zmin,item->extents.zmax,
	   item->extents.avg);
  }
#endif
  if (group->childs->len==0) {
    group->extents.xmin = 0;
    group->extents.xmax = 0;
    group->extents.ymin = 0;
    group->extents.ymax = 0;
    group->extents.zmin = 0;
    group->extents.zmax = 0;
    group->extents.avg = 0;
  } else {
    item=g_array_index(group->childs,Gtk3dcanvasItem*,0);
    group->extents.xmin = item->extents.xmin;
    group->extents.xmax = item->extents.xmax;
    group->extents.ymin = item->extents.ymin;
    group->extents.ymax = item->extents.ymax;
    group->extents.zmin = item->extents.zmin;
    group->extents.zmax = item->extents.zmax;
    for (i=1; i<group->childs->len; i++) {
      item=g_array_index(group->childs,Gtk3dcanvasItem*,i);
      group->extents.xmin = MIN(group->extents.xmin, 
				 item->extents.xmin);
      group->extents.xmax = MAX(group->extents.xmax, 
				 item->extents.xmax);
      group->extents.ymin = MIN(group->extents.ymin, 
				 item->extents.ymin);
      group->extents.ymax = MAX(group->extents.ymax, 
				 item->extents.ymax);
      group->extents.zmin = MIN(group->extents.zmin, 
				 item->extents.zmin);
      group->extents.zmax = MAX(group->extents.zmax, 
				 item->extents.zmax);
    }
    
    group->extents.avg = ((group->extents.xmax - group->extents.xmin) + 
			  (group->extents.ymax - group->extents.ymin) + 
			  (group->extents.zmax - group->extents.zmin)) / 3;
    
  }/* num_child != 0 */
#ifdef DEBUG
  printf("gtk_3dcanvas_group_compute_extend():x:%.2f-%.2f y:%.2f-%.2f z:%.2f-%.2f avg:%.2f\n",
	 group->extents.xmin,group->extents.xmax,
	 group->extents.ymin,group->extents.ymax,
	 group->extents.zmin,group->extents.zmax,
	 group->extents.avg);
#endif
}

void gtk_3dcanvas_group_add_child(Gtk3dcanvasGroup *group,
				  Gtk3dcanvasItem *item) {
  item->group=group;
  g_array_append_val(group->childs,item);
  item->canvas = GTK_3DCANVAS_ITEM(item->group)->canvas;
}

void gtk_3dcanvas_group_remove_child(Gtk3dcanvasGroup *group,
				     Gtk3dcanvasItem *item) {
  int i;

  item->group=NULL;
  for (i=0;
       i<group->childs->len 
	 && item != g_array_index(group->childs,Gtk3dcanvasItem*,i);
       i++);
  if (i == i<group->childs->len) {
    printf("Damnit! Can't remove an item from a group if it's not in it\n");
    return;
  }
  g_array_remove_index(group->childs,i);
}

/**
 * gtk_3dcanvas_group_dump
 * @group: what to dump
 * @verbose: if true, dump the contained items also
 *
 * Dump an item to screen for debugging purpose
 **/
void gtk_3dcanvas_group_dump(Gtk3dcanvasGroup* group,int verbose) {

  int i;
  Gtk3dcanvasItem *child;

  if (!group) {
    printf("This group is NULL\n");
    return;
  }
  printf ("Group %p contains %d childs (%p).\n",
	  group,group->childs->len,group->childs->data);
  printf("  extents=x:%.2f-%.2f y:%.2f-%.2f z:%.2f-%.2f avg:%.2f\n",
	 group->extents.xmin,group->extents.xmax,
	 group->extents.ymin,group->extents.ymax,
	 group->extents.zmin,group->extents.zmax,
	 group->extents.avg);
  if (verbose) {
    for (i=0; i<group->childs->len; i++) {
      child=GTK_3DCANVAS_ITEM(g_array_index(group->childs,
					    Gtk3dcanvasGroup*,
					    i));
      if (IS_GTK_3DCANVAS_GROUP(child)) {
	gtk_3dcanvas_group_dump(GTK_3DCANVAS_GROUP(child),
				  verbose);
      } else {
	gtk_3dcanvas_item_dump(child);
      }
    }
  }
}
