/* gtk-3dcanvas-pyramid - a pyramid to put on a gtk-3dcanvas */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/* Useful */
#define PI                      3.14159265358979323846264338327
#define DEG(angleRad)           ((angleRad) * 180 / PI)
#define RAD(angle360)           ((angle360) * PI / 180)
#define SQR(x)                  ((x) * (x))
#define SIGN(x)                 (((x) >= 0) ? 1.0 : -1.0)

/*---------------------------------\
| Property and signals definitions |
\---------------------------------*/
enum {
        PYRAMID_PROP_0,
	PYRAMID_PROP_SHAPE_TYPE,
	PYRAMID_PROP_PHI,
	PYRAMID_PROP_TOP_X,
	PYRAMID_PROP_TOP_Y,
	PYRAMID_PROP_TOP_Z,
	PYRAMID_PROP_BOTTOM_X,
	PYRAMID_PROP_BOTTOM_Y,
	PYRAMID_PROP_BOTTOM_Z,
	PYRAMID_PROP_RADIUS,
	PYRAMID_PROP_SMOTHNESS
};
/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_pyramid_class_init (Gtk3dcanvasPyramidClass *class);
static void gtk_3dcanvas_pyramid_init(Gtk3dcanvasPyramid *glb);

static GObject *
gtk_3dcanvas_pyramid_constructor (GType             type,
			       guint             n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_pyramid_finalize (GObject *object);
static void
gtk_3dcanvas_pyramid_dispose (GObject *object);

static void
gtk_3dcanvas_pyramid_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec);
static void
gtk_3dcanvas_pyramid_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
                                GParamSpec           *pspec);

static void
_gtk_3dcanvas_pyramid_add_pyramid(Gtk3dcanvasPyramid *pyramid);

static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_pyramid_get_type]-----------------------------------------*/
GType gtk_3dcanvas_pyramid_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasPyramidClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_pyramid_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasPyramid),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_pyramid_init,
    };

    type = g_type_register_static (gtk_3dcanvas_item_get_type (),
				   "Gtk3dcanvasPyramid",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_pyramid_class_init ]------------------------------------*/
static void gtk_3dcanvas_pyramid_class_init (Gtk3dcanvasPyramidClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->constructor  = gtk_3dcanvas_pyramid_constructor;
  object_class->set_property = gtk_3dcanvas_pyramid_set_property;
  object_class->get_property = gtk_3dcanvas_pyramid_get_property;
  object_class->finalize     = gtk_3dcanvas_pyramid_finalize;
  object_class->dispose      = gtk_3dcanvas_pyramid_dispose;


  g_object_class_install_property
    (object_class, PYRAMID_PROP_SHAPE_TYPE,
     g_param_spec_enum ("draw_type", 
			 _("Draw type"), 
			 _("The type of base of the pyramid. (Either cylindric or cubic)"),
			GTK_3DCANVAS_TYPE_3DCANVAS_SHAPE_TYPE,
			GTK_3DCANVAS_SHAPE_TYPE_BOX,
			(G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, PYRAMID_PROP_PHI,
     g_param_spec_float ("phi", 
			 _("Phi"), 
			 _("The rotation applied to the base points over the central axis (in radian)"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, PYRAMID_PROP_TOP_X,
     g_param_spec_float ("top_x", 
			 _("X top"), 
			 _("X coordinate of the top point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, PYRAMID_PROP_TOP_Y,
     g_param_spec_float ("top_y", 
			 _("Y top"), 
			 _("Y coordinate of the top point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, PYRAMID_PROP_TOP_Z,
     g_param_spec_float ("top_z", 
			 _("Z top"), 
			 _("Z coordinate of the top point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, PYRAMID_PROP_BOTTOM_X,
     g_param_spec_float ("bottom_x", 
			 _("X bottom"), 
			 _("X coordinate of the bottom point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, PYRAMID_PROP_BOTTOM_Y,
     g_param_spec_float ("bottom_y", 
			 _("Y bottom"), 
			 _("Y coordinate of the bottom point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, PYRAMID_PROP_BOTTOM_Z,
     g_param_spec_float ("bottom_z", 
			 _("Z bottom"), 
			 _("Z coordinate of the bottom point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, PYRAMID_PROP_RADIUS,
     g_param_spec_float ("radius", 
			 _("Radius"), 
			 _("Radius of the pyramid"),
			 0,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, PYRAMID_PROP_SMOTHNESS,
     g_param_spec_int ("smoothness", 
		       _("Smoothness"), 
		       _("Smoothness of the pyramid. The bigger it is, the more points the pyramid will contain. About 10 is really ok."),
		       0,G_MAXINT,0,
		       (G_PARAM_READWRITE)));

}

/*----[ gl_canvas_pyramid_constructor ]--------------------------------------*/
static GObject *
gtk_3dcanvas_pyramid_constructor (GType             type,
				guint            n_construct_properties,
			  GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_pyramid_finalize ]--------------------------------------*/
static void
gtk_3dcanvas_pyramid_finalize (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_pyramid_dispose ]------------------------------------------*/
static void
gtk_3dcanvas_pyramid_dispose (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->dispose (object);
}

/*----[ gtk_3dcanvas_pyramid_init ]---------------------------------------------*/
static void gtk_3dcanvas_pyramid_init(Gtk3dcanvasPyramid *pyramid) {
  pyramid->base_shape = GTK_3DCANVAS_SHAPE_TYPE_BOX;
} 

/*----[ gtk_3dcanvas_pyramid_set_property ]-------------------------------------*/
static void
gtk_3dcanvas_pyramid_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec) {
  Gtk3dcanvasPyramid *pyramid;
  Gtk3dcanvasItem *item;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_PYRAMID (object));

  pyramid = GTK_3DCANVAS_PYRAMID(object);
  item = GTK_3DCANVAS_ITEM (object);

  switch (param_id) {
  case PYRAMID_PROP_PHI:
    pyramid->phi = g_value_get_float(value);
    //    printf("phi=%.1f\n",pyramid->phi);
    while (pyramid->phi>PI) pyramid->phi -= 2*PI;
    while (pyramid->phi<-PI) pyramid->phi += 2*PI;
    //    printf("phi=%.1f\n",pyramid->phi);
    break;
  case PYRAMID_PROP_TOP_X:
    pyramid->top.x = g_value_get_float(value);
    printf("top_x=%.1f\n",pyramid->top.x);
    break;
  case PYRAMID_PROP_TOP_Y:
    pyramid->top.y = g_value_get_float(value);
    printf("top_y=%.1f\n",pyramid->top.y);
    break;
  case PYRAMID_PROP_TOP_Z:
    pyramid->top.z = g_value_get_float(value);
    printf("top_z=%.1f\n",pyramid->top.z);
    break;

  case PYRAMID_PROP_BOTTOM_X:
    pyramid->bottom.x = g_value_get_float(value);
    printf("bottom_x=%.1f\n",pyramid->bottom.x);
    break;
  case PYRAMID_PROP_BOTTOM_Y:
    pyramid->bottom.y = g_value_get_float(value);
    printf("bottom_y=%.1f\n",pyramid->bottom.y);
    break;
  case PYRAMID_PROP_BOTTOM_Z:
    pyramid->bottom.z = g_value_get_float(value);
    printf("bottom_z=%.1f\n",pyramid->bottom.z);
    break;

  case PYRAMID_PROP_RADIUS:
    pyramid->radius=g_value_get_float(value);
    printf("radius=%.1f\n",pyramid->radius);
    break;
  case PYRAMID_PROP_SMOTHNESS:
    pyramid->smoothness=g_value_get_int(value);
    break;

  default:
    printf("Property not known\n");
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
  gtk_3dcanvas_item_empty(item);
  _gtk_3dcanvas_pyramid_add_pyramid(pyramid);
  gtk_3dcanvas_draw(item->canvas);
}

/*----[ gtk_3dcanvas_pyramid_get_property ]----------------------------------*/
static void
gtk_3dcanvas_pyramid_get_property (GObject              *object,
				   guint                 param_id,
				   GValue               *value,
				   GParamSpec           *pspec) {
  Gtk3dcanvasItem *item;
  Gtk3dcanvasPyramid *pyramid;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_PYRAMID (object));

  pyramid = GTK_3DCANVAS_PYRAMID(object);
  item = GTK_3DCANVAS_ITEM (object);

  switch (param_id) {
  case PYRAMID_PROP_PHI:
    g_value_set_float (value, pyramid->phi);
    break;
  case PYRAMID_PROP_TOP_X:
    g_value_set_float (value, pyramid->top.x);
    break;
  case PYRAMID_PROP_TOP_Y:
    g_value_set_float (value, pyramid->top.y);
    break;
  case PYRAMID_PROP_TOP_Z:
    g_value_set_float (value, pyramid->top.z);
    break;

  case PYRAMID_PROP_BOTTOM_X:
    g_value_set_float (value, pyramid->bottom.x);
    break;
  case PYRAMID_PROP_BOTTOM_Y:
    g_value_set_float (value, pyramid->bottom.y);
    break;
  case PYRAMID_PROP_BOTTOM_Z:
    g_value_set_float (value, pyramid->bottom.z);
    break;

  case PYRAMID_PROP_RADIUS:
    g_value_set_float (value, pyramid->radius);
    break;

  case PYRAMID_PROP_SMOTHNESS:
    g_value_set_int (value, pyramid->smoothness);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
}

/*----[ _gtk_3dcanvas_pyramid_add_pyramid ]----------------------------------*/
static void
_gtk_3dcanvas_pyramid_add_pyramid(Gtk3dcanvasPyramid *pyramid) {
  
  float x0,y0,z0,x1,y1,z1;
  float radius;
  int smoothness;
  Gtk3dcanvasItem *item;

  float phi,phidelta;
  float x, y, z; /* coord of the point to add */
  float hyp; /* length of vector top-bottom */
  int nb_points, point;
  int begin_v=0, bottom_v, top_v = 0;
  int obj_v;

  transfo_t *rot; /* to make the base normal to the bot-top axis */

  g_return_if_fail (pyramid != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_PYRAMID (pyramid));
  
  item=GTK_3DCANVAS_ITEM(pyramid);
  x0 =pyramid->top.x;
  y0 =pyramid->top.y;
  z0 =pyramid->top.z;
  x1 =pyramid->bottom.x;
  y1 =pyramid->bottom.y;
  z1 =pyramid->bottom.z;
  radius = pyramid->radius;
  smoothness = pyramid->smoothness;
  phidelta = pyramid->phi;

  switch (pyramid->base_shape) {
  case GTK_3DCANVAS_SHAPE_TYPE_NONE:
    return;

  case GTK_3DCANVAS_SHAPE_TYPE_TRIANGLE:
    nb_points=3;
    break;

  case GTK_3DCANVAS_SHAPE_TYPE_BOX:
    nb_points=4;
    break;

  case GTK_3DCANVAS_SHAPE_TYPE_CYLINDER:
    nb_points=pyramid->smoothness * 2;
    break;

  default:
    printf("shape not known in _gtk_3dcanvas_pyramid_add_pyramid\n");
    return;
  }

#ifdef DEBUG
  printf("Add pyramid %.1f %.1f %.1f -> %.1f %.1f %.1f. Radius=%.1f %d pts\n",
  	 x0,y0,z0,x1,y1,z1,radius,nb_points);
#endif

  /* Preliminar computations to rotate the base so that (bottom->top) is normal to it */
  hyp=sqrt(SQR(x1-x0) + SQR(y1-y0) + SQR(z1-z0));
  if (hyp<0.00001) {
    hyp=0.00001;
  }
  rot = gtk_3dcanvas_transfo_new();
  //  printf("empty\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_x(rot, acos((x1-x0)/hyp));
  //  printf("rotx\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_y(rot, acos((y1-y0)/hyp));
  //  printf("roty\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_z(rot, acos((z1-z0)/hyp));
  //  printf("rotz\n");gtk_3dcanvas_transfo_dump(rot);
  
  /* bottom vertex */
  bottom_v = gtk_3dcanvas_item_add_point( item, x0, y0, z0, 0.0, 0.0, -1.0 );
  
  /* top vertex */
  top_v = gtk_3dcanvas_item_add_point( item, x1, y1, z1, 0.0, 0.0, 1.0 );
  gtk_3dcanvas_item_add_index( item, top_v );

  /* base vertices */
  for (point = 0; point < nb_points; point++) {
    phi = RAD(360.0 * (float)point / (float)nb_points) + phidelta;
    while (phi>PI)  phi -= 2*PI;
    while (phi<-PI) phi += 2*PI;
    x = 0;
    y = radius * cos( phi );
    z = radius * sin( phi );

    //    printf("(phi=%.2f (%.2f %.2f %.2f) -> ",phi,x,y,z);
    gtk_3dcanvas_transfo_apply(rot,&x,&y,&z);
    //    printf("(%.2f %.2f %.2f)\n",x,y,z);

    obj_v = gtk_3dcanvas_item_add_point( item, x0+x, y0+y, z0+z, x, y, z );
    gtk_3dcanvas_item_add_index( item, top_v );
    gtk_3dcanvas_item_add_index( item, obj_v );
    if (point == 0)
      begin_v = obj_v;
  }
  
  /* Complete the pyramid */
  gtk_3dcanvas_item_add_index( item, begin_v );
  gtk_3dcanvas_item_add_index( item, top_v );
  gtk_3dcanvas_item_add_index( item, begin_v );

}
/*-----------------\
| Public functions |
\-----------------*/
/**
 * gtk_3dcanvas_pyramid_new:
 * @parent: group in which to add the pyramid
 * @color: the color of the stick.
 * @first_arg_name: A list of object argument name/value pairs, 
 * NULL-terminated, used to configure the stick. For example:
 * "type",STICK_TYPE_CYLINDER, "top_x",1, "top_y",1, "top_z",1, NULL
 * @Varargs: Rest of this list.
 * @Returns: the newly created pyramid.
 * 
 * Make a pyramid item. 
 **/
Gtk3dcanvasItem* gtk_3dcanvas_pyramid_new(Gtk3dcanvasGroup *parent,
					  char *color,
					  const char *first_arg_name,...) {

  Gtk3dcanvasPyramid *pyramid;
  Gtk3dcanvasItem *item;
  va_list args;
  
  g_return_val_if_fail (IS_GTK_3DCANVAS_GROUP (parent), NULL);

  item=gtk_3dcanvas_item_new(parent,GTK_3DCANVAS_PYRAMID_TYPE,
			     "type",GL_QUAD_STRIP,
			     "color",color,
			     NULL);
  pyramid=GTK_3DCANVAS_PYRAMID(item);

  /* Update the properties */
  va_start (args, first_arg_name);
  g_object_set_valist (G_OBJECT (item), first_arg_name, args);
  va_end (args);

  /* add the pyramid */
  _gtk_3dcanvas_pyramid_add_pyramid(pyramid);
  return item;
}
