/* gtk-3dcanvas-stick - a stick to put on a gtk-3dcanvas */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/* Useful */
#define PI                      3.14159265358979323846264338327
#define DEG(angleRad)           ((angleRad) * 180 / PI)
#define RAD(angle360)           ((angle360) * PI / 180)
#define SQR(x)                  ((x) * (x))
#define SIGN(x)                 (((x) >= 0) ? 1.0 : -1.0)

/*---------------------------------\
| Property and signals definitions |
\---------------------------------*/
enum {
        STICK_PROP_0,
	STICK_PROP_BEGIN_X,
	STICK_PROP_BEGIN_Y,
	STICK_PROP_BEGIN_Z,

	STICK_PROP_END_X,
	STICK_PROP_END_Y,
	STICK_PROP_END_Z,

	STICK_PROP_PHI,
	STICK_PROP_SHAPE_TYPE,
	STICK_PROP_RADIUS,
	STICK_PROP_SMOTHNESS
};
/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_stick_class_init (Gtk3dcanvasStickClass *class);
static void gtk_3dcanvas_stick_init(Gtk3dcanvasStick *glb);

static GObject *
gtk_3dcanvas_stick_constructor (GType             type,
			       guint             n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_stick_finalize (GObject *object);
static void
gtk_3dcanvas_stick_dispose (GObject *object);

static void
gtk_3dcanvas_stick_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec);
static void
gtk_3dcanvas_stick_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
                                GParamSpec           *pspec);

static void
_gtk_3dcanvas_stick_add_stick(Gtk3dcanvasStick *stick);

static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_stick_get_type]-----------------------------------------*/
GType gtk_3dcanvas_stick_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasStickClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_stick_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasStick),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_stick_init,
    };

    type = g_type_register_static (gtk_3dcanvas_item_get_type (),
				   "Gtk3dcanvasStick",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_stick_class_init ]------------------------------------*/
static void gtk_3dcanvas_stick_class_init (Gtk3dcanvasStickClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->constructor  = gtk_3dcanvas_stick_constructor;
  object_class->set_property = gtk_3dcanvas_stick_set_property;
  object_class->get_property = gtk_3dcanvas_stick_get_property;
  object_class->finalize     = gtk_3dcanvas_stick_finalize;
  object_class->dispose      = gtk_3dcanvas_stick_dispose;

  g_object_class_install_property
    (object_class, STICK_PROP_BEGIN_X,
     g_param_spec_float ("begin_x", 
			 _("Begin X"), 
			 _("X coordinate of the begin point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_BEGIN_Y,
     g_param_spec_float ("begin_y", 
			 _("Begin Y"), 
			 _("Y coordinate of the begin point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_BEGIN_Z,
     g_param_spec_float ("begin_z", 
			 _("Begin Z"), 
			 _("Z coordinate of the begin point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, STICK_PROP_END_X,
     g_param_spec_float ("end_x", 
			 _("End X"), 
			 _("X coordinate of the end point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_END_Y,
     g_param_spec_float ("end_y", 
			 _("End Y"), 
			 _("Y coordinate of the end point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_END_Z,
     g_param_spec_float ("end_z", 
			 _("End Z"), 
			 _("Z coordinate of the end point"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, STICK_PROP_SHAPE_TYPE,
     g_param_spec_enum ("draw_type", 
			 _("Draw type"), 
			 _("The type of base of the stick. (Either cylindric or cubic)"),
			GTK_3DCANVAS_TYPE_3DCANVAS_SHAPE_TYPE,
			GTK_3DCANVAS_SHAPE_TYPE_BOX,
			(G_PARAM_READWRITE)));

  g_object_class_install_property
    (object_class, STICK_PROP_PHI,
     g_param_spec_float ("phi", 
			 _("Phi"), 
			 _("The rotation applied to the base points over the central axis (in radian)"),
			 -G_MAXFLOAT,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_RADIUS,
     g_param_spec_float ("radius", 
			 _("Radius"), 
			 _("Radius of the stick"),
			 0,G_MAXFLOAT,0,
			 (G_PARAM_READWRITE)));
  g_object_class_install_property
    (object_class, STICK_PROP_SMOTHNESS,
     g_param_spec_int ("smoothness", 
		       _("Smoothness"), 
		       _("Smoothness of the stick. The bigger it is, the more points the stick will contain. About 10 is really ok."),
		       0,G_MAXINT,0,
		       (G_PARAM_READWRITE)));

}

/*----[ gl_canvas_stick_constructor ]--------------------------------------*/
static GObject *
gtk_3dcanvas_stick_constructor (GType             type,
				guint            n_construct_properties,
			  GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_stick_finalize ]--------------------------------------*/
static void
gtk_3dcanvas_stick_finalize (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_stick_dispose ]------------------------------------------*/
static void
gtk_3dcanvas_stick_dispose (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->dispose (object);
}

/*----[ gtk_3dcanvas_stick_init ]---------------------------------------------*/
static void gtk_3dcanvas_stick_init(Gtk3dcanvasStick *stick) {
  stick->base_shape = GTK_3DCANVAS_SHAPE_TYPE_BOX;
} 

/*----[ gtk_3dcanvas_stick_set_property ]-------------------------------------*/
static void
gtk_3dcanvas_stick_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec) {
  Gtk3dcanvasStick *stick;
  Gtk3dcanvasItem *item;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_STICK (object));

  stick = GTK_3DCANVAS_STICK(object);
  item = GTK_3DCANVAS_ITEM (object);
  
  /* FIXME: could be done with transformation on points without freeing anything */ 
  switch (param_id) {
  case STICK_PROP_PHI:
    stick->phi = g_value_get_float(value);
    while (stick->phi>PI)  stick->phi -= 2*PI;
    while (stick->phi<-PI) stick->phi += 2*PI;
    break;
    
  case STICK_PROP_BEGIN_X:
    stick->begin.x = g_value_get_float(value);
    break;
  case STICK_PROP_BEGIN_Y:
    stick->begin.y = g_value_get_float(value);
    break;
  case STICK_PROP_BEGIN_Z:
    stick->begin.z = g_value_get_float(value);
    break;

  case STICK_PROP_END_X:
    stick->end.x = g_value_get_float(value);
    break;
  case STICK_PROP_END_Y:
    stick->end.y = g_value_get_float(value);
    break;
  case STICK_PROP_END_Z:
    stick->end.z = g_value_get_float(value);
    break;

  case STICK_PROP_RADIUS:
    stick->radius=g_value_get_float(value);
    break;
  case STICK_PROP_SMOTHNESS:
    stick->smoothness=g_value_get_int(value);
    break;
    /* FIXME: implement setting the shape */

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
  
  gtk_3dcanvas_item_empty(item);
  _gtk_3dcanvas_stick_add_stick(stick);
  gtk_3dcanvas_draw(item->canvas);
}

/*----[ gtk_3dcanvas_stick_get_property ]------------------------------------*/
static void
gtk_3dcanvas_stick_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
				GParamSpec           *pspec) {
  Gtk3dcanvasItem *item;
  Gtk3dcanvasStick *stick;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_STICK (object));

  stick = GTK_3DCANVAS_STICK(object);
  item = GTK_3DCANVAS_ITEM (object);

  switch (param_id) {
  case STICK_PROP_BEGIN_X:
    g_value_set_float (value, stick->begin.x);
    break;
  case STICK_PROP_BEGIN_Y:
    g_value_set_float (value, stick->begin.y);
    break;
  case STICK_PROP_BEGIN_Z:
    g_value_set_float (value, stick->begin.z);
    break;

  case STICK_PROP_END_X:
    g_value_set_float (value, stick->end.x);
    break;
  case STICK_PROP_END_Y:
    g_value_set_float (value, stick->end.y);
    break;
  case STICK_PROP_END_Z:
    g_value_set_float (value, stick->end.z);
    break;

  case STICK_PROP_PHI:
    g_value_set_float (value, stick->phi);
    break;
  case STICK_PROP_RADIUS:
    g_value_set_float (value, stick->radius);
    break;
  case STICK_PROP_SMOTHNESS:
    g_value_set_int (value, stick->smoothness);
    break;
    /* FIXME: implement the get_property(shape) */

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
}

/*----[ _gtk_3dcanvas_stick_add_stick ]----------------------------------------*/
static void
_gtk_3dcanvas_stick_add_stick(Gtk3dcanvasStick *stick) {
  
  float x0,y0,z0;
  float x1,y1,z1;
  float radius;
  int smoothness;
  Gtk3dcanvasItem *item;

  float phi,phidelta;
  float x, y, z;
  float hyp; /* length of vector top-bottom */
  int nb_points, point;
  int bot_v=0, top_v = 0;
  int first_bot_v=0, first_top_v = 0;

  transfo_t *rot; /* to make the base normal to the bot-top axis */

  g_return_if_fail (stick != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_STICK (stick));
  
  item=GTK_3DCANVAS_ITEM(stick);
  x0 =stick->begin.x;
  y0 =stick->begin.y;
  z0 =stick->begin.z;
  x1 =stick->end.x;
  y1 =stick->end.y;
  z1 =stick->end.z;
  radius = stick->radius;
  smoothness = stick->smoothness;
  phidelta = stick->phi;

  switch (stick->base_shape) {
  case GTK_3DCANVAS_SHAPE_TYPE_NONE:
    return;

  case GTK_3DCANVAS_SHAPE_TYPE_TRIANGLE:
    nb_points=3;
    break;

  case GTK_3DCANVAS_SHAPE_TYPE_BOX:
    nb_points=4;
    break;

  case GTK_3DCANVAS_SHAPE_TYPE_CYLINDER:
    nb_points=stick->smoothness * 2;
    break;

  default:
    printf("shape not known in _gtk_3dcanvas_pyramid_add_stick\n");
    return;
  }

#ifdef DEBUG
  printf("Add stick: (%.1f %.1f %.1f) -> (%.1f %.1f %.1f); radius=%.1f; smooth=%d; %d points\n",
	 x0,y0,z0,x1,y1,z1,radius,smoothness,nb_points);
#endif

  /* Preliminar computations to rotate the base so that (bottom->top) is normal to it */
  hyp=sqrt(SQR(x1-x0) + SQR(y1-y0) + SQR(z1-z0));
  if (hyp<0.00001) {
    hyp=0.00001;
  }
  rot = gtk_3dcanvas_transfo_new();
  //  printf("empty\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_x(rot, acos((x1-x0)/hyp));
  //  printf("rotx\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_y(rot, acos((y1-y0)/hyp));
  //  printf("roty\n");gtk_3dcanvas_transfo_dump(rot);
  gtk_3dcanvas_transfo_rot_z(rot, acos((z1-z0)/hyp));
  //  printf("rotz\n");gtk_3dcanvas_transfo_dump(rot);

  /* base vertices */
  for (point = 0; point < nb_points; point++) {
    phi = RAD(360.0 * (float)point / (float)nb_points) + phidelta;
    while (phi>PI)  phi -= 2*PI;
    while (phi<-PI) phi += 2*PI;
    x = 0;
    y = radius * cos( phi );
    z = radius * sin( phi );

    //    printf("(phi=%.2f (%.2f %.2f %.2f) -> ",phi,x,y,z);
    gtk_3dcanvas_transfo_apply(rot,&x,&y,&z);
    //    printf("(%.2f %.2f %.2f)\n",x,y,z);

    top_v = gtk_3dcanvas_item_add_point( item, x0+x, y0+y, z0+z, x, y, z );
    bot_v = gtk_3dcanvas_item_add_point( item, x1+x, y1+y, z1+z, x, y, z );
    gtk_3dcanvas_item_add_index( item, top_v );
    gtk_3dcanvas_item_add_index( item, bot_v );
    if (point == 0) {
      first_top_v = top_v;
      first_bot_v = bot_v;
    }
  }
  
  /* Complete the stick */
  gtk_3dcanvas_item_add_index( item, first_top_v );
  gtk_3dcanvas_item_add_index( item, first_bot_v );

}

/*-----------------\
| Public functions |
\-----------------*/
/**
 * gtk_3dcanvas_stick_new:
 * @parent: group in which to add the stick
 * @color: the color of the stick.
 * @first_arg_name: A list of object argument name/value pairs, 
 * NULL-terminated, used to configure the stick. For example:
 * "type",STICK_TYPE_CYLINDER, "begin_x",1, "begin_y",1, "begin_z",1, NULL
 * @Varargs: Rest of this list.
 * @Returns: the newly created stick.
 * 
 * Make a stick item. 
 **/
Gtk3dcanvasItem* gtk_3dcanvas_stick_new(Gtk3dcanvasGroup *parent,
					char *color,
					const char *first_arg_name,...) {

  Gtk3dcanvasItem *item;
  va_list args;

  g_return_val_if_fail (IS_GTK_3DCANVAS_GROUP (parent), NULL);

  item=gtk_3dcanvas_item_new(parent,GTK_3DCANVAS_STICK_TYPE,
			     "type",GL_QUAD_STRIP,
			     "color",color,
			     NULL);

  /* Update the properties */
  va_start (args, first_arg_name);
  g_object_set_valist (G_OBJECT (item), first_arg_name, args);
  va_end (args);

  /* add the stick */
  _gtk_3dcanvas_stick_add_stick(GTK_3DCANVAS_STICK(item));
  return item;
}
