/* gtk-3dcanvas - A 3d canvas with advanced features like gl camera */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <gtk/gtk.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_class_init (Gtk3dcanvasClass *class);
static void gtk_3dcanvas_init(Gtk3dcanvas *glb);

static GObject *
gtk_3dcanvas_constructor (GType                  type,
			  guint                  n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_finalize (GObject *object);


static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_get_type]---------------------------------------------*/
GType gtk_3dcanvas_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvas),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_init,
    };

    type = g_type_register_static (GTK_TYPE_OBJECT,
				   "Gtk3dcanvas",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_class_init ]------------------------------------------*/
static void gtk_3dcanvas_class_init (Gtk3dcanvasClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  /* nothing to init in the class for now */
  object_class->constructor  = gtk_3dcanvas_constructor;
  //  object_class->set_property = gl_browser_set_property;
  //  object_class->get_property = gl_browser_get_property;
  object_class->finalize     = gtk_3dcanvas_finalize;

}

/*----[ gl_canvas_constructor ]----------------------------------------------*/
static GObject *
gtk_3dcanvas_constructor (GType                  type,
			  guint                  n_construct_properties,
			  GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_finalize ]---------------------------------------------*/
static void
gtk_3dcanvas_finalize (GObject *object)
{
  //  Gtk3dcanvas *glb = GTK_3DCANVAS (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_init ]-------------------------------------------------*/
static void gtk_3dcanvas_init(Gtk3dcanvas *can) {
  can->root=GTK_3DCANVAS_GROUP(g_object_new(GTK_3DCANVAS_GROUP_TYPE, NULL));
  GTK_3DCANVAS_ITEM(can->root)->canvas = can;

  can->default_mouse_sens = 0.5;
  can->default_phi = 195.0;
  can->default_theta = -10.0;
  can->default_fov = 60.0;

} 

/**
 * gtk_3dcanvas_root:
 * @canvas:  A #Gtk3dcanvas.
 *
 * Queries the root group of a canvas.
 *
 * Return value:  The root group of the specified canvas.
 */
Gtk3dcanvasGroup* gtk_3dcanvas_root         (Gtk3dcanvas *canvas) {
  return canvas->root;
}

/**
 * gtk_3dcanvas_clear:
 * @canvas: A #Gtk3dcanvas.
 *
 * empties the world of all objects
 **/
void gtk_3dcanvas_clear(Gtk3dcanvas *canvas){
  printf("FIXME: gtk_3dcanvas_clear(): not implemented\n");
}

/**
 * gtk_3dcanvas_draw:
 * @canvas: A #Gtk3dcanvas.
 *
 * Ask for a redraw of the scene.
 */
void              gtk_3dcanvas_draw         (Gtk3dcanvas *canvas) {
  int i;
  if (!canvas) return;
  for (i=0; i<canvas->camera_count; i++)
    gtk_3dcanvas_cameragl_draw(canvas->camera[i]);
}
