/* test_gtk3dcanvas - a proof of concept of the gtk-3dcanvas */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h"

#define VIEW_ASPECT 1.3

static const gint config_attributes[] = {
  GDK_GL_DOUBLEBUFFER,
  GDK_GL_RGBA,
  GDK_GL_RED_SIZE,        1,
  GDK_GL_GREEN_SIZE,      1,
  GDK_GL_BLUE_SIZE,       1,
  GDK_GL_DEPTH_SIZE,      12,
  GDK_GL_ATTRIB_LIST_NONE
};

Gtk3dcanvasItem *ball1,*ball2,*ball3,*pyramid,*stick;
Gtk3dcanvas *world;

/*****[ Callbacks ]***********************************************************/
gint on_window_destroy(GtkWidget *widget);
void idle_anim(gpointer ignored); 

gint on_window_destroy(GtkWidget *widget) {
  gtk_main_quit();
  return TRUE;
}

void idle_anim(gpointer ignored) {
  static float x=0;
  static float delta = .1;

  x += delta;
  if (x>=2 || x <= -2) delta = -delta;
  g_object_set(GTK_3DCANVAS_BALL(ball1),"center_x",x*5,NULL);
  g_object_set(GTK_3DCANVAS_BALL(ball2),"center_x",x*(-5),NULL);
  g_object_set(GTK_3DCANVAS_BALL(ball3),"center_x",x*5,NULL);
  g_object_set(GTK_3DCANVAS_PYRAMID(pyramid),"phi",x,NULL);
  g_object_set(GTK_3DCANVAS_STICK(stick),"phi",x,NULL);
}

/*****[ Main ]****************************************************************/
int main(int argc, char *argv[]) {

  GtkWidget   *window,*vbox,*button_quit;
  GtkWidget   *camera;
  GdkGLConfig *glconfig = NULL;
  
  /* initialize gtk */
  gtk_init( &argc, &argv );

  /* Set the glconfig up */
  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new (&config_attributes[0]);
  if (glconfig == NULL) {
    g_print ("*** Cannot find the double-buffered visual.\n");
    g_print ("*** Trying single-buffered visual.\n");
    
    /* Try single-buffered visual */
    glconfig = gdk_gl_config_new (&config_attributes[1]);
    if (glconfig == NULL) {
      g_print ("*** Cannot find an OpenGL-capable visual\n");
      exit (1);
    }
  }

  /**** create new top level window ****/
  window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "test gtk_gl_canvas");
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);
  g_signal_connect (G_OBJECT(window), "destroy",
                    G_CALLBACK(on_window_destroy), NULL);

  /* Create a vbox, so that we can put a Quit button */
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  /* Create the canvas */
  world = GTK_3DCANVAS(g_object_new(GTK_3DCANVAS_TYPE, NULL));

  /* Populate the canvas */

  ball1=gtk_3dcanvas_ball_new(gtk_3dcanvas_root(world),
			      0, 0, 0,
			      "blue",
			      0.125 /*radius*/, 10 /*smoothness*/);
  ball2=gtk_3dcanvas_ball_new(gtk_3dcanvas_root(world),
			      0, 2, 0,
			      "red",
			      0.1, 10);
  ball3=gtk_3dcanvas_ball_new(gtk_3dcanvas_root(world),
			      0, 0, 2,
			      "green",
			      0.1, 10);
  pyramid=gtk_3dcanvas_pyramid_new(gtk_3dcanvas_root(world),
				   "green",
				   "radius",1.0,
				   "top_x",3.0,
				   "top_y",1.5,
				   "top_z",1.5,
				   "bottom_x",0.0,
				   "bottom_y",0.0,
				   "bottom_z",0.0,
				   "phi",314.159,
				   NULL);
  stick=gtk_3dcanvas_stick_new(gtk_3dcanvas_root(world),
			       "red",
			       "radius",1.0,
			       "begin_x",0.0,
			       "begin_y",3.5,
			       "begin_z",0.5,
			       "end_x",0.0,
			       "end_y",-1.0,
			       "end_z",00.0,
			       "phi",314.159,
			       NULL);
  //  gtk_3dcanvas_group_dump(gtk_3dcanvas_root(world),1);

  /* Add a camera */
  camera = gtk_3dcanvas_cameragl_new(world,glconfig,NULL);
  gtk_widget_set_size_request(camera, 200,200/VIEW_ASPECT);

  gtk_box_pack_start (GTK_BOX (vbox), camera , TRUE, TRUE, 0);
  gtk_widget_show(camera);

  /* Create the quit button, and add it to windows */
  button_quit = gtk_button_new_with_label ("Quit");
  gtk_box_pack_start (GTK_BOX (vbox), button_quit, FALSE, FALSE, 0);
  
  g_signal_connect (G_OBJECT (button_quit), "clicked",
                    G_CALLBACK (on_window_destroy), NULL);
  gtk_widget_show(button_quit);

  /* Add the animation function */
  g_idle_add_full(GDK_PRIORITY_REDRAW,
  		  (GtkFunction) idle_anim,
  		  NULL,NULL);
  gtk_widget_show(window);
  gtk_main();

  return 0;
}

